import os
import logging
import pty
import fcntl
import struct
import termios
import asyncio
import subprocess
import re
import json
import select
from typing import Optional
from channels.generic.websocket import AsyncWebsocketConsumer
from asgiref.sync import sync_to_async
from django.utils import timezone
# 

# Configure logging
logger = logging.getLogger(__name__)


class StatsConsumer(AsyncWebsocketConsumer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.process = None
        self.master_fd = None
        self.stream_task = None
        self.keep_running = False
        self.config_is_connected = False
        self.config_connected_at = None

        logger.info("StatsConsumer initialized for config-based uptime.")

    @sync_to_async
    def _get_config_data(self):
        from api.models import Config
        # Fetches is_connected and connected_at from the Config model
        config = Config.objects.first()
        if config:
            return config.is_connected, config.connected_at
        return False, None

    async def connect(self):
        await self.accept()
        logger.info("WebSocket connection established.")
        self.keep_running = True

        # Create a PTY for KelVPNCli process
        self.master_fd, slave_fd = pty.openpty()

        # Set virtual terminal size (required for correct connection time output)
        desired_width = 200
        desired_height = 24
        fcntl.ioctl(
            slave_fd,
            termios.TIOCSWINSZ,
            struct.pack("HHHH", desired_height, desired_width, 0, 0),
        )

        # Start KelVPNCli -T as a subprocess
        self.process = subprocess.Popen(
            ["/opt/kelvpn/bin/KelVPNCli", "-T"],
            stdout=slave_fd,
            stderr=slave_fd,
            close_fds=True,
        )
        os.close(slave_fd)
        logger.info("KelVPNCli process started successfully.")

        # Set master_fd to non-blocking mode
        flags = fcntl.fcntl(self.master_fd, fcntl.F_GETFL)
        fcntl.fcntl(self.master_fd, fcntl.F_SETFL, flags | os.O_NONBLOCK)

        # Start background task to read statistics
        self.stream_task = asyncio.create_task(self.stream_stats())

    async def disconnect(self, close_code):
        logger.info(
            f"WebSocket closed (code: {close_code}); stopping KelVPNCli process"
        )
        self.keep_running = False

        # Cancel background task
        if self.stream_task:
            self.stream_task.cancel()

        # Terminate KelVPNCli process
        if self.process:
            self.process.terminate()
            self.process.wait()
            logger.info("KelVPNCli process terminated.")

        # Close PTY file descriptor
        if self.master_fd is not None:
            try:
                os.close(self.master_fd)
                self.master_fd = None
                logger.info("PTY master file descriptor closed.")
            except OSError as e:
                logger.warning(f"Error closing PTY master file descriptor: {str(e)}")

    async def stream_stats(self):
        """
        Main loop: every second, check for new data from PTY.
        If available, read the data, extract statistics using the new async
        parse_stats, and send them via WebSocket.
        """
        try:
            while self.keep_running:
                await asyncio.sleep(1)
                if not self.master_fd:
                    logger.warning("master_fd is None, cannot read stats.")
                    continue # Or break, depending on desired recovery
                
                ready, _, _ = select.select([self.master_fd], [], [], 0)
                if not ready:
                    continue

                try:
                    data = os.read(self.master_fd, 4096)
                except OSError as e:
                    logger.warning(f"Error reading from PTY: {e}")
                    continue
                if not data:
                    logger.info("No data from PTY; KelVPNCli may have exited.")
                    break  # Exit stream_stats if PTY stream ends

                decoded_data = data.decode("utf-8", errors="replace")
                lines = decoded_data.splitlines()

                for line in lines:
                    line = line.strip()
                    if not line:
                        continue

                    if "State: Disconnect" in line:
                        logger.warning(
                            "KelVPN state 'Disconnect'; closing WebSocket"
                        )
                        try:
                            await self.mark_config_disconnected()
                        except Exception as e:
                            logger.error(
                                f"Failed to update config for 'Disconnect' state: {e}"
                            )

                        try:
                            await self.send(
                                text_data=json.dumps(
                                    {
                                        "event": "disconnect",
                                        "message": "KelVPN is disconnected",
                                    }
                                )
                            )
                        except Exception as e:
                            logger.error(
                                f"Error sending disconnect event to WebSocket: {str(e)}"
                            )
                        self.keep_running = False
                        break  # Exit inner loop
                    
                    # parse_stats now handles filtering of non-stats lines.
                    logger.debug(f"Processing line for stats: {line}")
                    stats = await self.parse_stats(line) 
                    
                    if stats:
                        try:
                            await self.send(text_data=json.dumps(stats))
                            logger.info(f"Sent stats: {stats}")
                        except Exception as e:
                            logger.error(f"Error sending WebSocket data: {str(e)}")
                            self.keep_running = False
                            break  # Exit inner loop
                
                if not self.keep_running:  # inner loop broken
                    break  # exit outer loop

        except asyncio.CancelledError:
            logger.info("Stream stats task was cancelled.")
        except Exception as e:
            logger.error(f"Unexpected error in stream_stats: {str(e)}", exc_info=True)
        finally:
            logger.info("Exiting stream_stats loop.")

    async def parse_stats(self, line: str) -> Optional[dict]:
        """
        Extract key statistics from the received line and uses Config for uptime.
        Returns None if the line isn't a recognized stats line.
        """
        if not (
            "uptime" in line.lower()
        ):
            return None

        if self.config_connected_at is None:
            is_connected_db, connected_at_db = await self._get_config_data()
            if connected_at_db:
                self.config_is_connected = is_connected_db
                self.config_connected_at = connected_at_db
                logger.info(f"DB: conn={self.config_is_connected}, at={self.config_connected_at}")
            else:
                logger.info("DB: No conn time yet. Uptime 0. Will retry.")
        
        download_match = re.search(r"download\s+([\d.]+\s+\w+)", line, re.IGNORECASE)
        upload_match = re.search(r"upload\s+([\d.]+\s+\w+)", line, re.IGNORECASE)
        received_match = re.search(
            r"([\d.]+\s+\w+)\s+bytes received", line, re.IGNORECASE
        )
        sent_match = re.search(r"([\d.]+\s+\w+)\s+bytes sent", line, re.IGNORECASE)
        packets_received_match = re.search(
            r"(\d+)\s+packets received", line, re.IGNORECASE
        )
        packets_sent_match = re.search(r"(\d+)\s+packets sent", line, re.IGNORECASE)

        current_uptime_str = "00:00:00"

        if self.config_is_connected and self.config_connected_at:
            now_utc = timezone.now()
            logger.info(
                f"Comparing times: now={now_utc} (tz: {now_utc.tzinfo}), "
                f"connected_at={self.config_connected_at} (tz: {self.config_connected_at.tzinfo})"
            )
            if now_utc > self.config_connected_at: 
                duration_seconds = int((now_utc - self.config_connected_at).total_seconds())
                duration_seconds = max(0, duration_seconds)

                hh = duration_seconds // 3600
                mm = (duration_seconds % 3600) // 60
                ss = duration_seconds % 60
                current_uptime_str = f"{hh:02}:{mm:02}:{ss:02}"
            else:
                logger.warning(f"DB conn_at ({self.config_connected_at}) > now. Uptime 0.")

        return {
            "download": download_match.group(1) if download_match else "0 B",
            "upload": upload_match.group(1) if upload_match else "0 B",
            "bytes_received": received_match.group(1) if received_match else "0 B",
            "bytes_sent": sent_match.group(1) if sent_match else "0 B",
            "packets_received": int(packets_received_match.group(1)) if packets_received_match else 0,
            "packets_sent": int(packets_sent_match.group(1)) if packets_sent_match else 0,
            "uptime": current_uptime_str,
        }

    @sync_to_async
    def mark_config_disconnected(self):
        from api.models import Config

        config_path = "/opt/kelvpn/share/KelVPNCheck.conf"

        try:
            config = Config.objects.first()
            if config:
                config.is_connected = False
                config.save()
                logger.info("Config updated: is_connected=False (disconnect detected)")
            else:
                logger.warning("No Config object found to mark as disconnected.")
        except Exception as e:
            logger.error(f"Failed to update Config model: {e}")

        try:
            with open(config_path, "r") as file:
                lines = file.readlines()

            new_lines = []
            found_connected = False

            for line in lines:
                if line.startswith("KELVPN_CONNECTED="):
                    new_lines.append("KELVPN_CONNECTED=false\n")
                    found_connected = True
                else:
                    new_lines.append(line)

            if not found_connected:
                new_lines.append("KELVPN_CONNECTED=false\n")

            with open(config_path, "w") as file:
                file.writelines(new_lines)

            logger.info("KelVPNCheck.conf updated: KELVPN_CONNECTED=false")

        except Exception as e:
            logger.error(f"Failed to update KelVPNCheck.conf: {e}")
